@@ -13,6 +13,8 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 
+from __future__ import absolute_import
+
 import logging
 
 from st2client.models import core
@@ -26,6 +28,7 @@ class RunnerType(core.Resource):
     _display_name = 'Runner'
     _plural = 'RunnerTypes'
     _plural_display_name = 'Runners'
+    _repr_attributes = ['name', 'enabled', 'description']
 
 
 class Action(core.Resource):
@@ -33,8 +36,17 @@ class Action(core.Resource):
     _repr_attributes = ['name', 'pack', 'enabled', 'runner_type']
 
 
-class ActionExecution(core.Resource):
+class Execution(core.Resource):
     _alias = 'Execution'
     _display_name = 'Action Execution'
+    _url_path = 'executions'
     _plural = 'ActionExecutions'
-    _plural_display_name = 'Action Executions'
+    _plural_display_name = 'Action executions'
+    _repr_attributes = ['status', 'action', 'start_timestamp', 'end_timestamp', 'parameters',
+                        'delay']
+
+
+# NOTE: LiveAction has been deprecated in favor of Execution. It will be left here for
+# backward compatibility reasons until v3.2.0
+class LiveAction(Execution):
+    pass
